@@ -58,7 +58,35 @@ enum {
 	OO_NS_XSD,
 	OO_NS_XSI,
 
-	OO_NS_PRESENT	/* added in gsf-1.14.8 */
+	OO_NS_PRESENT,	/* added in gsf-1.14.8 */
+
+	/* new in 3.0 */
+	OO_NS_RPT,
+	OO_NS_OF,
+	OO_NS_RDFA,
+	OO_NS_FIELD,
+	OO_NS_FORMX,
+
+	/* Other OpenDocument 1.1 */
+	OO_NS_ANIM,
+	OO_NS_DATASTYLE,
+	OO_NS_MANIFEST,
+	OO_NS_SMIL,
+
+	/* Symphony 1.3 */
+	OO_LOTUS_NS_PRODTOOLS,
+
+	/* KOffice 1.6.3 */
+	OO_KDE_NS_KOFFICE,
+
+	/*CleverAge ODF Add-in for Microsoft Office 3.0.5224.0 (11.0.8302)*/
+	OO_CLEVERAGE_NS_DC,
+
+	/* Microsoft Excel Formulas */
+	OO_MS_NS_MSOXL,
+
+	/* Gnumeric ODF extensions */
+	OO_GNUM_NS_EXT
 };
 
 extern GsfXMLInNS gsf_ooo_ns[];
